from xorq.common.utils import classproperty


try:
    from enum import StrEnum
except ImportError:
    from strenum import StrEnum


default_branch = "main"


class InitTemplates(StrEnum):
    cached_fetcher = "cached-fetcher"
    sklearn = "sklearn"
    penguins = "penguins"

    @classproperty
    def default(self):
        return self.cached_fetcher

    def get_default_branch(template, default_branch=default_branch):
        return dict(templates_branches).get(template, default_branch)


# NOTE: These are commit hashes from when the template update occurred
templates_branches = (
    (InitTemplates.cached_fetcher, "832543bcee23a6db0f835ddee64364583e156a94"),
    (InitTemplates.sklearn, "64f8a2b03d582773a1bfa5879c2befba6d23f675"),
    (InitTemplates.penguins, "56680c2fdb5745ae1a9c45394f28cf3707e26670"),
)



import pytest

import xorq.api as xo
import xorq.expr.datatypes as dt
from xorq.common.utils.graph_utils import to_node
from xorq.common.utils.lineage_utils import (
    ColorScheme,
    GenericNode,
    _build_column_tree,
    build_column_trees,
    build_tree,
)


@xo.udf.make_pandas_udf(
    schema=xo.schema({"price": float, "discount": float}),
    return_type=dt.float,
    name="calculate_discount_value",
)
def calculate_discount_value(df):
    return df["price"] * df["discount"]


@pytest.fixture
def sample_expression():
    sales_table = xo.memtable(
        {
            "order_id": [1, 2, 1, 2],
            "price": [100.0, 150.0, 200.0, 250.0],
            "discount": [0.1, 0.2, 0.15, 0.1],
        },
        name="sales",
    )

    sales_with_discount = sales_table.mutate(
        discount_value=calculate_discount_value.on_expr(sales_table)
    )

    expr = sales_with_discount.group_by("order_id").agg(
        total_discount=xo._.discount_value.sum(),
        total_price=xo._.price.sum(),
    )

    return expr


def test__build_column_tree_basic(sample_expression):
    node = to_node(sample_expression)

    lineage_tree = _build_column_tree(node)

    assert isinstance(lineage_tree, GenericNode)
    assert lineage_tree.op is not None
    assert isinstance(lineage_tree.children, tuple)

    assert len(lineage_tree.children) > 0

    for child in lineage_tree.children:
        assert isinstance(child, GenericNode)


def test_build_column_trees_and_display(sample_expression):
    column_trees = build_column_trees(sample_expression)

    assert "total_discount" in column_trees
    assert "total_price" in column_trees

    total_discount_tree = column_trees["total_discount"]
    total_price_tree = column_trees["total_price"]

    assert isinstance(total_discount_tree, GenericNode)
    assert isinstance(total_price_tree, GenericNode)

    display_tree = build_tree(total_discount_tree, dedup=True, max_depth=5)

    from rich.tree import Tree

    assert isinstance(display_tree, Tree)

    custom_palette = ColorScheme()
    display_tree_custom = build_tree(
        total_discount_tree, palette=custom_palette, dedup=False
    )
    assert isinstance(display_tree_custom, Tree)


def test_complete_lineage_for_total_discount_column(sample_expression):
    import xorq.vendor.ibis.expr.operations as ops
    from xorq.vendor.ibis.expr.operations.reductions import Sum

    column_trees = build_column_trees(sample_expression)
    total_discount_tree = column_trees["total_discount"]

    def gen_lineage_nodes(node):
        yield node.op
        for child in node.children:
            yield from gen_lineage_nodes(child)

    lineage_nodes = list(gen_lineage_nodes(total_discount_tree))

    # Expected lineage based on actual structure:
    # Sum -> Field(discount_value) -> UDF(calculate_discount_value) -> Field(price/discount) -> InMemoryTable

    node_types = [type(node).__name__ for node in lineage_nodes]

    sum_nodes = [node for node in lineage_nodes if isinstance(node, Sum)]
    assert len(sum_nodes) > 0, (
        f"Should contain Sum operation, found types: {node_types}"
    )

    udf_nodes = [
        node
        for node in lineage_nodes
        if type(node).__name__ == "calculate_discount_value"
    ]
    assert len(udf_nodes) > 0, (
        f"Should contain calculate_discount_value UDF in lineage, found types: {node_types}"
    )

    field_nodes = [node for node in lineage_nodes if isinstance(node, ops.Field)]
    field_names = [node.name for node in field_nodes]

    assert "discount_value" in field_names, (
        f"Should find discount_value field, found: {field_names}"
    )

    source_fields = {"price", "discount"}
    found_source_fields = source_fields.intersection(set(field_names))
    assert len(found_source_fields) > 0, (
        f"Should find at least one source field {source_fields}, found: {field_names}"
    )

    table_nodes = [
        node
        for node in lineage_nodes
        if isinstance(node, (ops.InMemoryTable, ops.UnboundTable, ops.DatabaseTable))
    ]
    assert len(table_nodes) > 0, "Should trace back to original table"



import operator

import pandas as pd
import pyarrow as pa
import pytest

import xorq.api as xo
from xorq.flight.exchanger import make_udxf
from xorq.flight.tests.conftest import (
    do_agg,
    field_name,
    my_udf,
    my_udf_on_expr,
    return_type,
)


def test_flight_expr(con, diamonds, baseline):
    unbound_expr = (
        xo.table(diamonds.schema()).pipe(do_agg).mutate(my_udf_on_expr).order_by("cut")
    )
    expr = xo.expr.relations.flight_expr(
        diamonds,
        unbound_expr,
        inner_name="flight-expr",
        name="remote-expr",
        con=con,
    )
    df = expr.execute()
    pd.testing.assert_frame_equal(
        baseline.sort_values("cut", ignore_index=True),
        df.sort_values("cut", ignore_index=True),
        check_exact=False,
    )


def test_flight_udxf(con, diamonds, baseline):
    input_expr = diamonds.pipe(do_agg)
    process_df = operator.methodcaller("assign", **{field_name: my_udf.fn})
    maybe_schema_in = input_expr.schema()
    maybe_schema_out = xo.schema(input_expr.schema() | {field_name: return_type})
    expr = xo.expr.relations.flight_udxf(
        input_expr,
        process_df=process_df,
        maybe_schema_in=maybe_schema_in,
        maybe_schema_out=maybe_schema_out,
        con=con,
        # operator.methodcaller doesn't have name, so must explicitly pass
        make_udxf_kwargs={"name": my_udf.__name__},
    ).order_by("cut")
    df = expr.execute()
    actual = df.sort_values("cut", ignore_index=True)
    expected = baseline.sort_values("cut", ignore_index=True)
    pd.testing.assert_frame_equal(
        actual,
        expected,
        check_exact=False,
    )


def test_make_udxf_fails():
    def dummy(df: pd.DataFrame):
        return pd.DataFrame({"row_count": [42]})

    with pytest.raises(ValueError):
        make_udxf(
            dummy,
            xo.schema({"dummy": "int64"}),
            pa.schema(
                [
                    ("row_count", pa.int64()),
                ]
            ),
        )

    with pytest.raises(ValueError):
        make_udxf(
            dummy,
            pa.schema(
                [
                    ("dummy", pa.int64()),
                ]
            ),
            xo.schema({"row_count": "int64"}),
        )



import pandas as pd
import pytest

import xorq.api as xo
from xorq.cli import lineage_command
from xorq.ibis_yaml.compiler import build_expr


def test_lineage_no_target(capsys):
    # When build target does not exist, exit with code 2 and print error
    with pytest.raises(SystemExit) as exc:
        lineage_command("nonexistent")
    assert exc.value.code == 2
    out = capsys.readouterr().out
    assert "Build target not found: nonexistent" in out


def test_lineage_simple(tmp_path, capsys):
    # Create a simple build of a mutate expression
    # Create an in-memory table and expression
    df = pd.DataFrame({"a": [1, 2], "b": [3, 4]})
    table = xo.memtable(df, name="tbl")
    expr = table.mutate(c=table["a"] + table["b"])
    # Compile expression to a build directory

    # build_expr writes expr.yaml, metadata.json, etc.
    build_path = build_expr(
        expr,
        builds_dir=tmp_path / "builds",
        cache_dir=tmp_path / "cache",
    )
    # Invoke lineage command
    lineage_command(str(build_path))
    out = capsys.readouterr().out
    # Expect lineage for each column
    # Check that lineage sections appear for each column
    assert "Lineage for column 'a':" in out
    assert "Field:a" in out
    assert "Lineage for column 'b':" in out
    assert "Field:b" in out
    assert "Lineage for column 'c':" in out
    # The root of 'c' lineage is an Add operation
    assert "Add #1" in out



from xorq.catalog import (
    CATALOG_YAML_FILENAME,
    XorqCatalog,
    load_catalog,
    resolve_build_dir,
)


def test_load_and_save_catalog(tmp_path):
    cat = load_catalog(path=str(tmp_path / CATALOG_YAML_FILENAME))
    assert hasattr(cat, "entries") and isinstance(cat.entries, tuple)
    cat.save(path=str(tmp_path / CATALOG_YAML_FILENAME))
    cat2 = load_catalog(path=str(tmp_path / CATALOG_YAML_FILENAME))
    assert cat2.entries == cat.entries


def test_resolve_target_alias_and_entry():
    catalog = XorqCatalog.from_dict(
        {
            "aliases": {"foo": {"entry_id": "e1", "revision_id": "r1"}},
            "entries": [
                {
                    "entry_id": "e1",
                    "current_revision": "r1",
                    "history": [{"revision_id": "r1"}],
                }
            ],
        }
    )
    t = catalog.resolve_target("foo")
    assert t.entry_id == "e1"
    assert t.rev == "r1"
    assert t.alias is True
    t2 = catalog.resolve_target("e1@r1")
    assert t2.entry_id == "e1"
    assert t2.rev == "r1"
    assert t2.alias is False


def test_resolve_build_dir_by_path(tmp_path):
    d = tmp_path / "x"
    d.mkdir()
    cat = {"aliases": {}, "entries": []}
    p = resolve_build_dir(str(d), cat)
    assert p == d


def test_resolve_build_dir_by_build_id(tmp_path):
    d = tmp_path / "build1"
    d.mkdir()
    entry = {
        "entry_id": "e1",
        "current_revision": "r1",
        "history": [{"revision_id": "r1", "build": {"build_id": "b1", "path": str(d)}}],
    }
    cat = XorqCatalog.from_dict({"aliases": {}, "entries": [entry]})
    p = resolve_build_dir("e1", cat)
    assert p == d



import functools
import json
import os
import shutil
import subprocess
import sys
import uuid
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, Mapping, Optional, Tuple, Union

import toolz
import yaml
from attrs import evolve, field, frozen
from attrs.validators import deep_iterable, instance_of, optional

import xorq as xo
from xorq.common.utils.dask_normalize.dask_normalize_utils import (
    file_digest,
)
from xorq.common.utils.func_utils import (
    if_not_none,
)
from xorq.ibis_yaml.compiler import (
    DumpFiles,
    load_expr,
)


CATALOG_YAML_FILENAME = "catalog.yaml"


def get_default_catalog_path():
    # dynamically retrieve: tests need to monkeypatch XDG_CONFIG_HOME
    return (
        Path(os.environ.get("XDG_CONFIG_HOME") or Path.home().joinpath(".config"))
        .joinpath("xorq", CATALOG_YAML_FILENAME)
        .absolute()
    )


def get_catalog_path(path: Optional[Union[str, Path]] = None) -> Path:
    """Return the catalog file path, using XDG_CONFIG_HOME if set or default to ~/.config."""
    return Path(path) if path else get_default_catalog_path()


get_now_utc = functools.partial(datetime.now, timezone.utc)


@toolz.curry
def to_dict(self, **kwargs):
    dct = self.__getstate__()
    dct = dct | {k: v(dct[k]) for k, v in kwargs.items()}
    return dct


@toolz.curry
def from_dict(cls, dct: dict, **kwargs):
    modifications = {k: v(dct.get(k)) for k, v in kwargs.items()}
    return cls(**dct | modifications)


def convert_uuid(value):
    match value:
        case None:
            return uuid.uuid4()
        case str():
            return uuid.UUID(value)
        case uuid.UUID():
            return value
        case _:
            raise ValueError


def convert_datetime(value):
    match value:
        case None:
            return get_now_utc()
        case str():
            return datetime.fromisoformat(value)
        case datetime():
            return value
        case _:
            raise ValueError


def convert_build(value):
    match value:
        case None:
            return None
        case dict():
            return Build.from_dict(value)
        case Build():
            return value
        case _:
            raise ValueError


@frozen
class CatalogMetadata:
    """Catalog metadata."""

    # FIXME: make uuid
    catalog_id: str = field(
        validator=instance_of(uuid.UUID), factory=uuid.uuid4, converter=convert_uuid
    )
    created_at: str = field(validator=instance_of(datetime), factory=get_now_utc)
    updated_at: str = field(validator=instance_of(datetime), factory=get_now_utc)
    tool_version: str = field(validator=instance_of(str), default=xo.__version__)

    def with_updated_timestamp(self) -> "CatalogMetadata":
        """Return new metadata with updated timestamp."""
        return self.evolve(updated_at=get_now_utc())

    evolve = evolve

    to_dict = to_dict(catalog_id=str)

    from_dict = classmethod(
        from_dict(created_at=convert_datetime, updated_at=convert_datetime)
    )


@frozen
class Build:
    """Build information."""

    build_id: Optional[str] = field(default=None, validator=optional(instance_of(str)))
    path: Optional[Path] = field(
        default=None,
        validator=optional(instance_of(Path)),
        converter=toolz.curried.excepts(Exception, Path),
    )

    evolve = evolve

    to_dict = to_dict(path=if_not_none(str))

    from_dict = classmethod(from_dict)


@frozen
class Revision:
    """Revision data."""

    # FIXME: make int
    revision_id: str = field(validator=instance_of(str))
    created_at: str = field(
        validator=instance_of(datetime),
        factory=get_now_utc,
        converter=convert_datetime,
    )
    build: Optional[Build] = field(
        default=None, validator=optional(instance_of(Build)), converter=convert_build
    )
    expr_hashes: Optional[Dict[str, str]] = field(default=None)
    node_hashes: Tuple[str, ...] = field(
        factory=tuple,
        validator=deep_iterable(instance_of(str), instance_of(tuple)),
        converter=tuple,
    )
    meta_digest: Optional[str] = field(
        default=None, validator=optional(instance_of(str))
    )
    metadata: Optional[Dict[str, Any]] = field(default=None)
    schema_fingerprint: Optional[str] = field(
        default=None, validator=optional(instance_of(str))
    )

    evolve = evolve

    to_dict = to_dict(build=if_not_none(Build.to_dict))

    from_dict = classmethod(from_dict)


@frozen
class Entry:
    """Entry in the catalog."""

    entry_id: str = field(validator=instance_of(str))
    current_revision: Optional[str] = field(
        default=None, validator=optional(instance_of(str))
    )
    history: Tuple[Revision, ...] = field(
        validator=deep_iterable(instance_of(Revision), instance_of(tuple)),
        factory=tuple,
    )

    @property
    def created_at(self):
        return min(
            (revision.created_at for revision in self.history),
            default=None,
        )

    def with_revision(self, revision: Revision) -> "Entry":
        """Add a new revision to history."""
        return self.evolve(
            current_revision=revision.revision_id, history=self.history + (revision,)
        )

    def maybe_get_revision(self, revision_id: str) -> Optional[Revision]:
        """Get revision by ID if it exists."""
        return next((r for r in self.history if r.revision_id == revision_id), None)

    def maybe_current_revision(self) -> Optional[Revision]:
        """Get current revision if it exists."""
        if self.current_revision is None:
            return None
        return self.maybe_get_revision(self.current_revision)

    evolve = evolve

    to_dict = to_dict(
        history=toolz.compose(tuple, functools.partial(map, Revision.to_dict))
    )

    @classmethod
    def from_dict(cls, dct):
        dct = dct.copy()
        created_at = dct.pop("created_at", None)
        obj = from_dict(
            cls,
            dct,
            history=toolz.compose(tuple, functools.partial(map, Revision.from_dict)),
        )
        if created_at is not None:
            assert datetime.fromisoformat(created_at) == min(
                rev.created_at for rev in obj.history
            )
        return obj


@frozen
class Alias:
    entry_id: str = field(validator=instance_of(str))
    revision_id: Optional[str] = field(
        default=None, validator=optional(instance_of(str))
    )
    updated_at: str = field(
        validator=optional(instance_of(datetime)),
        default=None,
        converter=convert_datetime,
    )

    evolve = evolve

    to_dict = to_dict

    from_dict = classmethod(from_dict)


@frozen
class XorqCatalog:
    """Xorq Catalog container."""

    aliases: Mapping[str, Alias] = field(factory=dict)
    entries: Tuple[Entry, ...] = field(
        validator=deep_iterable(instance_of(Entry), instance_of(tuple)),
        factory=tuple,
        converter=tuple,
    )
    api_version: str = field(default="xorq.dev/v1", validator=instance_of(str))
    kind: str = field(default="XorqCatalog", validator=instance_of(str))
    metadata: CatalogMetadata = field(
        factory=CatalogMetadata,
        validator=optional(instance_of(CatalogMetadata)),
    )

    def __attrs_post_init__(self):
        if self.metadata is None:
            object.__setattr__(self, "metadata", CatalogMetadata())

    def with_entry(self, entry: Entry) -> "XorqCatalog":
        """Add or update an entry."""
        existing_entries = tuple(
            e for e in self.entries if e.entry_id != entry.entry_id
        )
        return self.evolve(entries=existing_entries + (entry,))

    def with_alias(self, name: str, alias: Alias) -> "XorqCatalog":
        """Add or update an alias."""
        return self.evolve(aliases=self.aliases | {name: alias})

    def maybe_get_entry(self, entry_id: str) -> Optional[Entry]:
        """Get entry by ID if it exists."""
        return next((e for e in self.entries if e.entry_id == entry_id), None)

    def maybe_get_alias(self, name: str) -> Optional[Alias]:
        """Get alias by name if it exists."""
        return self.aliases.get(name)

    def maybe_get_revision_by_token(self, token: str) -> Optional[Revision]:
        """Get revision by entry@revision token if it exists."""
        t = self.resolve_target(token)
        if t is None:
            return None
        gen = (
            rev
            for entry in self.entries
            for rev in entry.history
            if entry.entry_id == t.entry_id and rev.revision_id == t.rev
        )
        return next(gen, None)

    def with_updated_metadata(self) -> "XorqCatalog":
        """Return catalog with updated timestamp."""
        return self.evolve(metadata=self.metadata.with_updated_timestamp())

    def get_entry_ids(self) -> Tuple[str, ...]:
        """Get all entry IDs."""
        return tuple(e.entry_id for e in self.entries)

    evolve = evolve

    def resolve_target(self, target: str):
        return Target.from_str(target, self)

    to_dict = to_dict(
        aliases=toolz.curried.valmap(Alias.to_dict),
        entries=toolz.compose(tuple, functools.partial(map, Entry.to_dict)),
        metadata=CatalogMetadata.to_dict,
    )

    def to_yaml(self, path):
        dct = self.to_dict()
        with path.open("wt") as fh:
            yaml.safe_dump(dct, fh, sort_keys=False)

    def save(self, path):
        catalog_path = get_catalog_path(path)
        catalog_path.parent.mkdir(parents=True, exist_ok=True)
        self.to_yaml(catalog_path)

    from_dict = classmethod(
        from_dict(
            aliases=toolz.curried.valmap(Alias.from_dict),
            entries=toolz.compose(tuple, functools.partial(map, Entry.from_dict)),
            metadata=if_not_none(CatalogMetadata.from_dict),
        )
    )

    @classmethod
    def from_path(cls, path):
        if path.exists():
            with Path(path).open() as fh:
                dct = yaml.safe_load(fh)
            return cls.from_dict(dct) if dct else cls()
        else:
            return cls()

    @classmethod
    def from_default(cls):
        pass


@frozen
class Target:
    entry_id: str = field(validator=instance_of(str))
    rev: Optional[str] = field(default=None, validator=optional(instance_of(str)))
    alias: bool = field(default=False, validator=instance_of(bool))

    @classmethod
    def from_str(cls, target: str, catalog: XorqCatalog = None):
        catalog = catalog or load_catalog()
        (base, rev) = target.split("@", 1) if "@" in target else (target, None)
        if alias := catalog.aliases.get(base):
            return Target(
                entry_id=alias.entry_id, rev=rev or alias.revision_id, alias=True
            )
        if entry := catalog.maybe_get_entry(base):
            return Target(entry_id=base, rev=rev or entry.current_revision, alias=False)
        return None


@frozen
class ServerRecord:
    pid: int = field(validator=instance_of(int))
    command: str = field(validator=instance_of(str))
    target: str = field(validator=instance_of(str))
    port: Optional[int] = field(validator=optional(instance_of(int)), default=None)
    start_time: datetime = field(
        validator=instance_of(datetime),
        factory=get_now_utc,
        converter=convert_datetime,
    )
    node_hash: Optional[str] = field(validator=optional(instance_of(str)), default=None)

    def clone(self, **changes) -> "ServerRecord":
        """Return a new ServerRecord with updated fields."""
        return evolve(self, **changes)

    to_dict = to_dict

    def to_json_dict(self):
        data = toolz.dissoc(
            self.to_dict(),
            "node_hash",
        )
        data = data | toolz.valfilter(
            bool,
            {
                "to_unbind_hash": self.node_hash,
                "start_time": self.start_time.isoformat(),
            },
        )
        return data

    def save(self, record_dir: Path) -> Path:
        """Side effect: save a ServerRecord to JSON file in record_dir."""
        record_dir.mkdir(parents=True, exist_ok=True)
        path = record_dir / f"{self.pid}.json"
        path.write_text(json.dumps(self.to_json_dict()))
        return path

    @property
    def running(self):
        try:
            os.kill(self.pid, 0)
            return True
        except Exception:
            return False

    from_dict = classmethod(from_dict)

    @classmethod
    def from_path(cls, path):
        return cls.from_dict(json.loads(path.read_text()))

    @classmethod
    def load_records(cls, record_dir: Path, only_running=True):
        if not record_dir.exists():
            return tuple()
        # FIXME: remove json files of records that aren't running
        records = tuple(
            map(toolz.excepts(Exception, cls.from_path), record_dir.glob("*.json"))
        )
        if only_running:
            records = tuple(record for record in records if record.running)
        return records


load_catalog = toolz.compose(XorqCatalog.from_path, get_catalog_path)


def resolve_build_dir(
    token: str, catalog: Optional[XorqCatalog] = None
) -> Optional[Path]:
    """Resolve build directory from raw catalog dict."""

    def absolutify(path):
        if not path.is_absolute():
            path = get_default_catalog_path().parent.joinpath(path)
        return path

    path = Path(token)
    if path.exists() and path.is_dir():
        return path
    revision = catalog.maybe_get_revision_by_token(token)
    if revision and revision.build and revision.build.path:
        return absolutify(revision.build.path)
    return None


def maybe_resolve_build_dirs(
    left: str, right: str, catalog
) -> tuple[Path, Path] | None:
    try:
        left_dir = resolve_build_dir(left, catalog)
        right_dir = resolve_build_dir(right, catalog)
        # Interpret relative build paths from catalog as relative to config directory
        config_path = get_catalog_path()
        config_dir = config_path.parent
        # Adjust left_dir if token is not a literal directory and left_dir is relative
        token_path = Path(left)
        if (
            not token_path.is_dir()
            and left_dir is not None
            and not left_dir.is_absolute()
        ):
            left_dir = config_dir / left_dir
        # Adjust right_dir similarly
        token_path = Path(right)
        if (
            not token_path.is_dir()
            and right_dir is not None
            and not right_dir.is_absolute()
        ):
            right_dir = config_dir / right_dir
    except ValueError as e:
        print(f"Error: {e}")
        return None
    if left_dir is None:
        print(f"Build target not found: {left}")
        return None
    if right_dir is None:
        print(f"Build target not found: {right}")
        return None
    if not left_dir.exists() or not left_dir.is_dir():
        print(f"Build directory not found: {left_dir}")
        return None
    if not right_dir.exists() or not right_dir.is_dir():
        print(f"Build directory not found: {right_dir}")
        return None
    return left_dir, right_dir


def get_diff_file_list(
    left_dir: Path, right_dir: Path, files: list[str] | None, all_flag: bool
) -> tuple[str, ...]:
    default = (DumpFiles.expr,)
    if files is not None:
        return tuple(files)
    if all_flag:
        default_files = (
            DumpFiles.expr,
            DumpFiles.deferred_reads,
            DumpFiles.profiles,
            DumpFiles.sql,
            DumpFiles.metadata,
        )
        sqls = {p.relative_to(left_dir).as_posix() for p in left_dir.rglob("*.sql")} | {
            p.relative_to(right_dir).as_posix() for p in right_dir.rglob("*.sql")
        }
        return default_files + tuple(sorted(sqls))
    return default


def do_diff_builds(
    left: str,
    right: str,
    files: list[str] | None,
    all_flag: bool,
) -> int:
    def get_keep_files(
        file_list: tuple[str, ...], left_dir: Path, right_dir: Path
    ) -> tuple[str, ...]:
        return tuple(
            f for f in file_list if (left_dir / f).exists() or (right_dir / f).exists()
        )

    def run_diffs(left_dir: Path, right_dir: Path, keep_files: tuple[str, ...]) -> int:
        exit_code = 0
        for f in keep_files:
            print(f"## Diff: {f}")
            lf = left_dir / f
            rf = right_dir / f
            ret = subprocess.call(["git", "diff", "--no-index", "--", str(lf), str(rf)])
            if ret == 1:
                exit_code = 1
            elif ret != 0:
                return ret
        return exit_code

    catalog = load_catalog()
    resolved = maybe_resolve_build_dirs(left, right, catalog)
    if not resolved:
        return 2
    left_dir, right_dir = resolved
    file_list = get_diff_file_list(left_dir, right_dir, files, all_flag)
    keep_files = get_keep_files(file_list, left_dir, right_dir)
    if not keep_files:
        print("No files to diff")
        return 2
    return run_diffs(left_dir, right_dir, keep_files)


def _get_catalog_paths() -> tuple[Path, Path]:
    """Return (config_path, config_dir) for the catalog."""
    config_path = get_catalog_path()
    return config_path, config_path.parent


def _do_unknown(args):
    print(f"Unknown catalog subcommand: {args.subcommand}")


@frozen
class AddBuildRequest:
    build_path: Path = field(converter=Path)
    alias: Optional[str] = field(default=None)

    def __attrs_post_init__(self):
        if not self.build_path.exists():
            raise ValueError(f"Build path does not exist: {self.build_path}")


@frozen
class BuildInfo:
    """Validated build information."""

    build_id: str = field()
    meta_digest: str = field()
    source_path: Path = field()


@frozen
class CatalogPaths:
    config_path: Path = field()
    config_dir: Path = field()
    builds_dir: Path = field()

    @classmethod
    def create(cls) -> "CatalogPaths":
        config_path, config_dir = _get_catalog_paths()
        builds_dir = config_dir / "catalog-builds"
        return cls(
            config_path=config_path, config_dir=config_dir, builds_dir=builds_dir
        )


@frozen
class AddBuildResult:
    entry_id: str = field()
    revision_id: str = field()
    build_id: str = field()


def validate_build(request: AddBuildRequest) -> BuildInfo:
    def get_meta_file(path):
        build_path = Path(path)
        meta_file = build_path / DumpFiles.metadata
        if not build_path.exists() or not build_path.is_dir():
            raise ValueError(f"Build path not found: {build_path}")
        if not meta_file.exists():
            raise ValueError(
                f"{DumpFiles.metadata} not found in build path: {build_path}"
            )
        return meta_file

    source_path = request.build_path.resolve()
    meta_file = get_meta_file(source_path)
    # The build_id is the directory name
    build_id = meta_file.parent.name
    # Compute meta_digest from metadata.json
    meta_digest = f"sha1:{file_digest(meta_file)}"

    return BuildInfo(
        build_id=build_id,
        meta_digest=meta_digest,
        source_path=source_path,
    )


def make_build_object(build_info: BuildInfo) -> Build:
    build_path_str = str(Path("catalog-builds") / build_info.build_id)
    return Build.from_dict({"build_id": build_info.build_id, "path": build_path_str})


def make_revision(
    build_info: BuildInfo,
    revision_id: str,
    timestamp: str,
) -> Revision:
    build_obj = make_build_object(build_info)
    revision_data = {
        "revision_id": revision_id,
        "created_at": timestamp,
        "build": build_obj,
        "meta_digest": build_info.meta_digest,
    }

    return Revision.from_dict(revision_data)


def maybe_find_existing_entry(catalog: XorqCatalog, alias: str) -> Optional[Entry]:
    if not alias:
        return None
    alias_obj = catalog.maybe_get_alias(alias)
    if not alias_obj:
        return None
    return catalog.maybe_get_entry(alias_obj.entry_id)


def compute_next_revision_id(entry: Entry) -> str:
    existing = [rev.revision_id for rev in entry.history]
    nums = [int(r[1:]) for r in existing if r.startswith("r") and r[1:].isdigit()]
    next_num = max(nums, default=0) + 1
    return f"r{next_num}"


def create_new_entry(build_info: BuildInfo, timestamp: str) -> Tuple[Entry, str]:
    entry_id = str(uuid.uuid4())
    revision_id = "r1"
    revision = make_revision(build_info, revision_id, timestamp)

    entry_data = {
        "entry_id": entry_id,
        "current_revision": revision_id,
        "history": (revision.to_dict(),),
    }
    entry = Entry.from_dict(entry_data)
    return entry, revision_id


def update_existing_entry(
    entry: Entry,
    build_info: BuildInfo,
    timestamp: str,
) -> Tuple[Entry, str]:
    revision_id = compute_next_revision_id(entry)
    revision = make_revision(build_info, revision_id, timestamp)
    updated_entry = entry.with_revision(revision)
    return updated_entry, revision_id


def update_catalog_with_entry(
    catalog: XorqCatalog,
    entry: Entry,
    revision_id: str,
    alias: Optional[str],
    timestamp: str,
) -> XorqCatalog:
    updated_catalog = catalog.with_entry(entry)

    if alias:
        alias_obj = Alias.from_dict(
            {
                "entry_id": entry.entry_id,
                "revision_id": revision_id,
                "updated_at": timestamp,
            }
        )
        updated_catalog = updated_catalog.with_alias(alias, alias_obj)

    return updated_catalog.with_updated_metadata()


def process_catalog_update(
    catalog: XorqCatalog,
    build_info: BuildInfo,
    alias: Optional[str],
    timestamp: str,
) -> Tuple[XorqCatalog, str, str]:
    existing_entry = maybe_find_existing_entry(catalog, alias)

    if existing_entry:
        entry, revision_id = update_existing_entry(
            existing_entry, build_info, timestamp
        )
    else:
        entry, revision_id = create_new_entry(build_info, timestamp)

    updated_catalog = update_catalog_with_entry(
        catalog, entry, revision_id, alias, timestamp
    )

    return updated_catalog, entry.entry_id, revision_id


def do_ensure_directories(paths: CatalogPaths) -> None:
    paths.config_dir.mkdir(parents=True, exist_ok=True)
    paths.builds_dir.mkdir(parents=True, exist_ok=True)


def do_copy_build_safely(build_info: BuildInfo, paths: CatalogPaths) -> None:
    target_dir = paths.builds_dir / build_info.build_id
    temp_dir = paths.builds_dir / f".{build_info.build_id}.tmp"

    if temp_dir.exists():
        shutil.rmtree(temp_dir)

    try:
        shutil.copytree(build_info.source_path, temp_dir)
        if target_dir.exists():
            shutil.rmtree(target_dir)
        os.replace(str(temp_dir), str(target_dir))
    except Exception:
        shutil.rmtree(temp_dir, ignore_errors=True)
        raise


def do_save_catalog(catalog: XorqCatalog, config_path: Path) -> None:
    catalog.save(path=config_path)


def do_print_result(result: AddBuildResult) -> None:
    print(
        f"Added build {result.build_id} as entry {result.entry_id} revision {result.revision_id}"
    )


def do_catalog_add(args) -> None:
    """Add a build to the catalog."""
    request = AddBuildRequest(build_path=args.build_path, alias=args.alias)
    paths = CatalogPaths.create()
    timestamp = get_now_utc().isoformat()

    build_info = validate_build(request)

    do_ensure_directories(paths)
    do_copy_build_safely(build_info, paths)

    catalog = load_catalog(path=paths.config_path)
    updated_catalog, entry_id, revision_id = process_catalog_update(
        catalog, build_info, request.alias, timestamp
    )

    do_save_catalog(updated_catalog, paths.config_path)
    result = AddBuildResult(
        entry_id=entry_id, revision_id=revision_id, build_id=build_info.build_id
    )
    do_print_result(result)


def do_catalog_ls(args):
    """List entries and aliases in the catalog."""
    config_path, _ = _get_catalog_paths()
    catalog = load_catalog(path=config_path)
    if aliases := catalog.aliases:
        print("Aliases:")
        print(
            "\n".join(
                f"{al}\t{mapping.entry_id}\t{mapping.revision_id}"
                for al, mapping in aliases.items()
            )
        )
    print("Entries:")
    for entry in catalog.entries:
        curr_rev = entry.current_revision
        build_id = None
        for rev in entry.history:
            if rev.revision_id == curr_rev:
                build_id = rev.build.build_id
                break
        print(f"{entry.entry_id}\t{curr_rev}\t{build_id}")


def compute_build_dir(
    build_path: Optional[Union[str, Path]], config_dir: Path
) -> Optional[Path]:
    if not build_path:
        return None
    build_dir = Path(build_path)
    if not build_dir.is_absolute():
        build_dir = config_dir / build_dir
    return build_dir


def do_catalog_info(args):
    config_path, _ = _get_catalog_paths()
    catalog = load_catalog(path=config_path)
    print(f"Catalog path: {config_path}")
    print(f"Entries: {len(catalog.entries)}")
    print(f"Aliases: {len(catalog.aliases)}")


def do_catalog_rm(args):
    config_path, _ = _get_catalog_paths()
    catalog = load_catalog(path=config_path)
    token = args.entry
    entry = next((e for e in catalog.entries if e.entry_id == token), None)
    if entry:
        entries = tuple(e for e in catalog.entries if e.entry_id != token)
        others = tuple(
            name for name, value in catalog.aliases.items() if value.entry_id == token
        )
        aliases = toolz.dissoc(catalog.aliases, *others)
        catalog = catalog.evolve(entries=entries, aliases=aliases)
        catalog.save(path=config_path)
        print(f"Removed entry {token}")
        return
    elif catalog.aliases.pop(token, None):
        catalog.save(path=config_path)
        print(f"Removed alias {token}")
        return
    else:
        print(f"Entry {token} not found in catalog")


def do_catalog_export(args):
    config_path, config_dir = _get_catalog_paths()
    export_dir = Path(args.output_path)
    if export_dir.exists() and not export_dir.is_dir():
        print(f"Export path exists and is not a directory: {export_dir}")
        return
    export_dir.mkdir(parents=True, exist_ok=True)
    if config_path.exists():
        shutil.copy2(config_path, export_dir / config_path.name)
    else:
        print(f"No catalog found at {config_path}")
        return
    src_builds = config_dir / "catalog-builds"
    if src_builds.exists() and src_builds.is_dir():
        dest_builds = export_dir / src_builds.name
        if dest_builds.exists():
            shutil.rmtree(dest_builds)
        shutil.copytree(src_builds, dest_builds)
    print(f"Exported catalog and builds to {export_dir}")


def do_catalog_diff_builds(args):
    code = do_diff_builds(args.left, args.right, args.files, args.all)
    sys.exit(code)


def lineage_command(
    target: str,
):
    """
    Print per-column lineage trees for a single build.
    """
    catalog = load_catalog()
    build_dir = resolve_build_dir(target, catalog)
    if build_dir is None or not build_dir.exists() or not build_dir.is_dir():
        print(f"Build target not found: {target}")
        sys.exit(2)
    # Load serialized expression
    expr = load_expr(build_dir)
    # Build and print lineage trees
    from xorq.common.utils.lineage_utils import build_column_trees, print_tree

    trees = build_column_trees(expr)
    for column, tree in trees.items():
        print(f"Lineage for column '{column}':")
        print_tree(tree)
        print()


_CATALOG_HANDLER_MAP = {
    "add": do_catalog_add,
    "ls": do_catalog_ls,
    "info": do_catalog_info,
    "rm": do_catalog_rm,
    "export": do_catalog_export,
    "diff-builds": do_catalog_diff_builds,
}


def _catalog_command_dispatch(args):
    return _CATALOG_HANDLER_MAP.get(args.subcommand, _do_unknown)(args)


catalog_command = _catalog_command_dispatch


def ps_command(cache_dir: str) -> None:
    """List active xorq servers."""

    def do_print_table(
        headers: Tuple[str, ...], rows: Tuple[Tuple[str, ...], ...]
    ) -> None:
        if rows:
            widths = [max(len(cell) for cell in col) for col in zip(headers, *rows)]
        else:
            widths = [len(h) for h in headers]
        fmt = "  ".join(f"{{:<{w}}}" for w in widths)
        print(fmt.format(*headers))
        for row in rows:
            print(fmt.format(*row))

    def format_server_table(
        records: Tuple[ServerRecord, ...],
    ) -> Tuple[Tuple[str, ...], Tuple[Tuple[str, ...], ...]]:
        headers = ("TARGET", "STATE", "COMMAND", "HASH", "PID", "PORT", "UPTIME")
        rows: list[tuple[str, ...]] = []
        now = get_now_utc()
        for rec in records:
            state = "running"
            try:
                delta = now - rec.start_time
                hours, rem = divmod(int(delta.total_seconds()), 3600)
                minutes, _ = divmod(rem, 60)
                uptime = f"{hours}h{minutes}m"
            except Exception:
                uptime = ""
            rows.append(
                (
                    rec.target,
                    state,
                    rec.command,
                    rec.node_hash or "",
                    str(rec.pid),
                    str(rec.port) if rec.port is not None else "",
                    uptime,
                )
            )
        return headers, tuple(rows)

    headers, rows = format_server_table(
        ServerRecord.load_records(Path(cache_dir) / "servers")
    )
    do_print_table(headers, rows)



import pytest

from xorq.cli import catalog_command, parse_args
from xorq.ibis_yaml.compiler import DumpFiles


LS_BASELINE = "Entries:\n"


@pytest.fixture(autouse=True)
def isolate_catalog(tmp_path, monkeypatch):
    """Redirect catalog file to a temporary location for each test."""
    # Point XDG_CONFIG_HOME to tmp_path so DEFAULT_CATALOG_PATH is under tmp_path
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path / "config"))
    # Ensure parent exists
    (tmp_path / "config").mkdir(parents=True, exist_ok=True)
    yield


def test_ls_empty_catalog(capsys):
    args = parse_args(["catalog", "ls"])
    catalog_command(args)
    out = capsys.readouterr().out
    # No aliases, just Entries header
    assert out == LS_BASELINE


def test_diff_builds_identical(tmp_path, capsys):
    # Create two identical build dirs
    for name in ("b1", "b2"):
        d = tmp_path / name
        d.mkdir()
        (d / DumpFiles.expr).write_text("foo: bar")
    # Run diff-builds
    args = parse_args(
        ["catalog", "diff-builds", str(tmp_path / "b1"), str(tmp_path / "b2")]
    )
    with pytest.raises(SystemExit) as e:
        catalog_command(args)
    # No differences -> exit code 0
    assert e.value.code == 0


def test_diff_builds_no_files(tmp_path):
    # Empty builds: no expr.yaml
    for name in ("b1", "b2"):
        (tmp_path / name).mkdir()
    args = parse_args(
        ["catalog", "diff-builds", str(tmp_path / "b1"), str(tmp_path / "b2")]
    )
    with pytest.raises(SystemExit) as e:
        catalog_command(args)
    # No files to diff -> exit code 2
    assert e.value.code == 2


def test_rm_entry(tmp_path, capsys):
    # Add and then remove a build entry
    build_dir = tmp_path / "b1"
    build_dir.mkdir()
    (build_dir / DumpFiles.metadata).write_text("{}")
    # Add build
    args = parse_args(["catalog", "add", str(build_dir)])
    catalog_command(args)
    out_add = capsys.readouterr().out
    entry_id = out_add.split()[5]
    # Remove entry
    args = parse_args(["catalog", "rm", entry_id])
    catalog_command(args)
    out_rm = capsys.readouterr().out
    assert f"Removed entry {entry_id}" in out_rm
    # ls should no longer list the entry
    args = parse_args(["catalog", "ls"])
    catalog_command(args)
    out_ls = capsys.readouterr().out
    assert entry_id not in out_ls


def test_rm_alias(tmp_path, capsys):
    # Add build with alias, then remove alias only
    build_dir = tmp_path / "b2"
    build_dir.mkdir()
    (build_dir / DumpFiles.metadata).write_text("{}")
    alias = "myalias"
    args = parse_args(["catalog", "add", "-a", alias, str(build_dir)])
    catalog_command(args)
    _ = capsys.readouterr()
    # Remove alias
    args = parse_args(["catalog", "rm", alias])
    catalog_command(args)
    out_rm = capsys.readouterr().out
    assert f"Removed alias {alias}" in out_rm
    # ls should not show any aliases
    args = parse_args(["catalog", "ls"])
    catalog_command(args)
    out_ls = capsys.readouterr().out
    assert "Aliases:" not in out_ls


def test_rm_not_found(tmp_path, capsys):
    # Attempt to remove non-existent entry
    args = parse_args(["catalog", "rm", "noexist"])
    catalog_command(args)
    out = capsys.readouterr().out
    assert "Entry noexist not found in catalog" in out


def test_info_empty_catalog(tmp_path, capsys):
    # On an empty catalog, info should show zero entries and aliases
    args = parse_args(["catalog", "info"])
    catalog_command(args)
    out = capsys.readouterr().out
    assert "Catalog path:" in out
    assert "Entries: 0" in out
    assert "Aliases: 0" in out


def test_info_after_add(tmp_path, capsys):
    # After adding one entry and one alias, info should update counts
    build_dir = tmp_path / "b1"
    build_dir.mkdir()
    (build_dir / DumpFiles.metadata).write_text("{}")
    # Add with alias
    alias = "foo"
    args = parse_args(["catalog", "add", "-a", alias, str(build_dir)])
    catalog_command(args)
    _ = capsys.readouterr()
    # Info now should report 1 entry and 1 alias
    args = parse_args(["catalog", "info"])
    catalog_command(args)
    out = capsys.readouterr().out
    assert "Entries: 1" in out
    assert "Aliases: 1" in out



from __future__ import annotations

from functools import lru_cache, singledispatch
from itertools import count
from typing import Any, Callable, Dict, Tuple

import dask.base
from attrs import evolve, field, frozen
from attrs.validators import instance_of
from rich import print as rprint
from rich.tree import Tree

import xorq.expr.relations as rel
import xorq.expr.udf as udf
import xorq.vendor.ibis.expr.operations as ops
from xorq.common.utils.graph_utils import (
    gen_children_of,
    to_node,
)
from xorq.vendor.ibis.expr.operations.core import Node


__all__ = [
    "build_column_trees",
    "build_tree",
    "print_tree",
]


@frozen
class GenericNode:
    op: Node = field(validator=instance_of(Node))
    children: Tuple["GenericNode", ...] = field(
        factory=tuple, validator=instance_of(tuple)
    )

    def map_children(
        self, fn: Callable[["GenericNode"], "GenericNode"]
    ) -> "GenericNode":
        return evolve(self, children=tuple(fn(c) for c in self.children))

    def clone(self, **changes: Any) -> "GenericNode":
        return evolve(self, **changes)


def _build_column_tree(node: Node) -> GenericNode:
    match node:
        case ops.Field(rel=ops.Project(values=values)) as field_node:
            # include the field and recurse into its mapped expression
            mapped = values[field_node.name]
            child = _build_column_tree(to_node(mapped))
            return GenericNode(op=field_node, children=(child,))

        case ops.Field() as field_node:
            children = tuple(
                _build_column_tree(to_node(child))
                for child in gen_children_of(field_node)
            )
            return GenericNode(op=field_node, children=children)

        case ops.Project() as proj:
            return _build_column_tree(to_node(proj.parent))

        case _:
            children = tuple(
                _build_column_tree(to_node(child)) for child in gen_children_of(node)
            )
            return GenericNode(op=node, children=children)


def build_column_trees(expr: Any) -> Dict[str, GenericNode]:
    """Builds a lineage tree for each column in the expression."""
    op = to_node(expr)
    cols = getattr(op, "values", None) or getattr(op, "fields", {})
    return {k: _build_column_tree(to_node(v)) for k, v in cols.items()}


@frozen
class ColorScheme:
    colors: Dict[str, str] = {
        "table": "[#658594]",  # dragonBlue2 (muted ocean blue)
        "cached_table": "[#8a739a]",  # dragonViolet (soft purple)
        "field": "[#8ea4a2]",  # dragonAsh (sage green)
        "literal": "[#b6927b]",  # dragonOrange2 (warm earth)
        "project": "[#c5c9c5]",  # dragonWhite (soft cloud)
        "filter": "[#87a987]",  # springViolet (forest green)
        "join": "[bold #a292a3]",  # springViolet2 (muted lavender)
        "aggregate": "[#c4b28a]",  # dragonYellow (wheat)
        "sort": "[#7d7c61]",  # comet (olive)
        "limit": "[#43436c]",  # dragonInk (deep twilight)
        "value": "[#b98d7b]",  # dragonOrange (clay)
        "binary": "[#7d957d]",  # dragonGreen2 (moss)
        "window": "[bold #8a739a]",  # bold dragonViolet
        "udf": "[#7e9cd8]",  # waveBlue1 (accent blue)
        "default": "[#a6a69c]",  # fujiGray (natural stone)
    }

    def get(self, category: str) -> str:
        return self.colors.get(category, self.colors["default"])


default_palette = ColorScheme()


def _category(node: Node) -> str:
    name_typs = (
        ops.Field,
        ops.Literal,
        ops.Project,
        ops.Filter,
        ops.Aggregate,
        ops.Sort,
        ops.Limit,
        ops.BinaryOp,
        ops.ValueOp,
    )
    if isinstance(node, name_typs):
        return node.__class__.__name__.lower()
    if isinstance(node, (ops.InMemoryTable, ops.UnboundTable, ops.DatabaseTable)):
        return "table"
    if isinstance(node, rel.RemoteTable):
        return "remote_table"
    if isinstance(node, rel.FlightExpr):
        return "flight"
    if isinstance(node, rel.FlightUDXF):
        return "udxf"
    if isinstance(node, udf.ExprScalarUDF):
        return "udf"
    if isinstance(node, rel.CachedNode):
        return "cached_table"
    if isinstance(node, rel.Read):
        return "table"
    if isinstance(node, ops.JoinChain):
        return "join"
    if isinstance(node, ops.WindowFunction):
        return "window"
    return "default"


@singledispatch
def format_node(node: Node, config: Dict[str, Any] | None = None) -> str:
    config = config or {}
    palette: ColorScheme = config.get("palette", default_palette)
    cat = _category(node)
    color = palette.get(cat)
    return f"{color}{node.__class__.__name__}[/]"


@format_node.register
def _(node: ops.Field, cfg: Dict[str, Any] | None = None) -> str:
    palette: ColorScheme = (cfg or {}).get("palette", default_palette)
    col = palette.get("field")
    return f"{col}Field:{node.name}[/]"


@format_node.register
def _(node: rel.RemoteTable, cfg: Dict[str, Any] | None = None) -> str:
    palette: ColorScheme = (cfg or {}).get("palette", default_palette)
    col = palette.get("remote_table")
    return f"{col}RemoteTable:{node.name}[/]"


@format_node.register
def _(node: rel.CachedNode, cfg: Dict[str, Any] | None = None) -> str:
    palette: ColorScheme = (cfg or {}).get("palette", default_palette)
    col = palette.get("cached_table")
    store = getattr(node.cache, "kind", "cache")
    return f"{col}Cache[{store}] {getattr(node, 'name', '')}[/]"


@format_node.register
def _(node: rel.FlightExpr, cfg: Dict[str, Any] | None = None) -> str:
    palette: ColorScheme = (cfg or {}).get("palette", default_palette)
    col = palette.get("flight")
    return f"{col}FlightExpr ({node.input_expr})[/]"


@format_node.register
def _(node: udf.ExprScalarUDF, cfg: Dict[str, Any] | None = None) -> str:
    palette: ColorScheme = (cfg or {}).get("palette", default_palette)
    col = palette.get("udxf")
    return f"{col}ExprScalarUDF[/]"


@format_node.register
def _(node: ops.WindowFunction, cfg: Dict[str, Any] | None = None) -> str:
    palette: ColorScheme = (cfg or {}).get("palette", default_palette)
    col = palette.get("window")

    parts = []
    if node.order_by:
        parts.append(f"order_by: {node.order_by}")
    if node.group_by:
        parts.append(f"group_by: {node.group_by}")
    if node.start:
        parts.append(f"start: {node.start}")
    if node.end:
        parts.append(f"end: {node.end}")

    if parts:
        details = "\n ".join(parts)
        return f"{col}WindowFunction:\n {details}[/]"
    else:
        return f"{col}WindowFunction[/]"


@format_node.register
def _(node: ops.Literal, cfg: Dict[str, Any] | None = None) -> str:
    palette: ColorScheme = (cfg or {}).get("palette", default_palette)
    col = palette.get("literal")
    return f"{col}Literal: {node.value}[/]"


@lru_cache
def _token_node(g: GenericNode) -> str:
    """unique token for the node based on its operation and children (useful in
    deduplication)"""
    op = g.op
    return dask.base.tokenize(
        (
            getattr(op, "name", None),  # is name always set?
            getattr(op, "schema", None),
            tuple(_token_node(c) for c in g.children),
        )
    )


def build_tree(
    node: GenericNode,
    *,
    palette: ColorScheme | None = None,
    dedup: bool = True,
    max_depth: int | None = None,
) -> Tree:
    cfg = {"palette": palette or default_palette}

    seen: dict[str, int] = {}
    seq = count(1)

    def _to_tree(g: GenericNode, depth: int) -> Tree:
        if max_depth is not None and depth > max_depth:
            return Tree("[dim]…[/]")

        digest = _token_node(g) if dedup else None
        if digest is not None and digest in seen:
            ref = seen[digest]
            return Tree(f"[italic dim]↻ see #{ref}[/]")

        ref = next(seq)
        if digest is not None:
            seen[digest] = ref

        label = format_node(g.op, cfg)
        if dedup:
            label += f" [grey37]#{ref}[/]"
        branch = Tree(label)

        for child in g.children:
            branch.add(_to_tree(child, depth + 1))
        return branch

    return _to_tree(node, 0)


def print_tree(
    node: GenericNode,
    *,
    palette: ColorScheme | None = None,
    dedup: bool = True,
    max_depth: int | None = None,
) -> None:
    rprint(build_tree(node, palette=palette, dedup=dedup, max_depth=max_depth))



import argparse
import os
import pdb
import sys
import traceback
from functools import partial
from pathlib import Path

from opentelemetry import trace

import xorq
import xorq.common.utils.pickle_utils  # noqa: F401
from xorq.caching.strategy import SnapshotStrategy
from xorq.catalog import (
    ServerRecord,
    catalog_command,
    lineage_command,
    ps_command,
    resolve_build_dir,
)
from xorq.common.utils import classproperty
from xorq.common.utils.caching_utils import get_xorq_cache_dir
from xorq.common.utils.import_utils import import_from_path
from xorq.common.utils.io_utils import maybe_open
from xorq.common.utils.logging_utils import get_print_logger
from xorq.common.utils.node_utils import expr_to_unbound
from xorq.common.utils.otel_utils import tracer
from xorq.flight import FlightServer
from xorq.ibis_yaml.compiler import (
    build_expr,
    load_expr,
)
from xorq.ibis_yaml.packager import (
    SdistBuilder,
    SdistRunner,
)
from xorq.init_templates import InitTemplates
from xorq.loader import load_backend
from xorq.vendor.ibis import Expr


try:
    from enum import StrEnum
except ImportError:
    from strenum import StrEnum


class OutputFormats(StrEnum):
    csv = "csv"
    json = "json"
    parquet = "parquet"
    arrow = "arrow"

    @classproperty
    def default(self):
        return self.parquet


logger = get_print_logger()


def ensure_build_dir(expr_path):
    build_dir = resolve_build_dir(expr_path)
    if build_dir is None or not build_dir.exists() or not build_dir.is_dir():
        print(f"Build target not found: {expr_path}")
        sys.exit(2)
    return build_dir


@tracer.start_as_current_span("cli.uv_build_command")
def uv_build_command(
    script_path,
    project_path=None,
    sys_argv=(),
):
    sdist_builder = SdistBuilder.from_script_path(
        script_path, project_path=project_path, args=sys_argv
    )
    # should we execv here instead?
    # ensure we do copy_sdist
    sdist_builder.build_path
    popened = sdist_builder._uv_tool_run_xorq_build
    print(popened.stderr, file=sys.stderr, end="")
    print(popened.stdout, file=sys.stdout, end="")
    return popened


@tracer.start_as_current_span("cli.uv_run_command")
def uv_run_command(
    expr_path,
    sys_argv=(),
):
    sdist_runner = SdistRunner(expr_path, args=sys_argv)
    popened = sdist_runner._uv_tool_run_xorq_run
    return popened


@tracer.start_as_current_span("cli.build_command")
def build_command(
    script_path,
    expr_name,
    builds_dir="builds",
    cache_dir=get_xorq_cache_dir(),
    debug: bool = False,
):
    """
    Generate artifacts from an expression in a given Python script

    Parameters
    ----------
    script_path : Path to the Python script
    expr_name : The name of the expression to build
    builds_dir : Directory where artifacts will be generated
    cache_dir : Directory where the parquet cache files will be generated

    Returns
    -------

    """

    span = trace.get_current_span()
    span.add_event(
        "build.params",
        {
            "script_path": str(script_path),
            "expr_name": expr_name,
        },
    )

    if not os.path.exists(script_path):
        raise ValueError(f"Error: Script not found at {script_path}")
    print(f"Building {expr_name} from {script_path}", file=sys.stderr)
    vars_module = import_from_path(script_path, module_name="__main__")
    match expr := getattr(vars_module, expr_name, None):
        case Expr():
            pass
        case None:
            raise ValueError(f"Expression {expr_name} not found")
        case _:
            raise ValueError(
                f"The object {expr_name} must be an instance of {Expr.__module__}.{Expr.__name__}"
            )

    build_path = build_expr(
        expr, builds_dir=builds_dir, cache_dir=Path(cache_dir), debug=debug
    )
    expr_hash = build_path.name
    span.add_event("build.outputs", {"expr_hash": expr_hash})
    print(
        f"Written '{expr_name}' to {build_path}",
        file=sys.stderr,
    )
    print(build_path)


@tracer.start_as_current_span("cli.run_command")
def run_command(
    expr_path,
    output_path=None,
    output_format=OutputFormats.default,
    cache_dir=get_xorq_cache_dir(),
    limit=None,
):
    """
    Execute an artifact

    Parameters
    ----------
    expr_path : str
        Path to the expr in the builds dir
    output_path : str
        Path to write output. Defaults to os.devnull
    output_format : OutputFormats | str, optional
        Output format, either "csv", "json", "arrow", or "parquet". Defaults to "parquet"
    cache_dir : Path, optional
        Directory where the parquet cache files will be generated
    limit : int, optional
        Limit number of rows to output. Defaults to None (no limit).

    Returns
    -------

    """

    span = trace.get_current_span()
    span.add_event(
        "run.params",
        {
            "expr_path": str(expr_path),
            "output_path": str(output_path),
            "output_format": output_format,
        },
    )

    expr = load_expr(expr_path, cache_dir=cache_dir)
    if limit is not None:
        expr = expr.limit(limit)
    arbitrate_output_format(expr, output_path, output_format)


def arbitrate_output_format(expr, output_path, output_format):
    match (output_path, output_format):
        case (None, _):
            output_path = os.devnull
        case ("-", OutputFormats.json):
            # FIXME: deal with windows
            output_path = sys.stdout
        case ("-", _):
            output_path = sys.stdout.buffer
        case _:
            pass
    match output_format:
        case OutputFormats.csv:
            expr.to_csv(output_path)
        case OutputFormats.json:
            expr.to_json(output_path)
        case OutputFormats.arrow:
            from xorq.expr.api import to_pyarrow_stream

            to_pyarrow_stream(expr, output_path)
        case OutputFormats.parquet:
            expr.to_parquet(output_path)
        case _:
            raise ValueError(f"Unknown output_format: {output_format}")


@tracer.start_as_current_span("cli.run_unbound_command")
def run_unbound_command(
    expr_path,
    to_unbind_hash=None,
    to_unbind_tag=None,
    output_path=None,
    output_format=OutputFormats.default,
    cache_dir=get_xorq_cache_dir(),
    limit=None,
    typ=None,
    instream=sys.stdin.buffer,
):
    """
    Execute an unbound expression by reading Arrow IPC from stdin and binding it

    Parameters
    ----------
    expr_path : str
        Path to the expr in the builds dir
    to_unbind_hash : str, optional
        Hash of the node to unbind
    to_unbind_tag : str, optional
        Tag of the node to unbind
    output_path : str, optional
        Path to write output. Defaults to stdout for arrow, os.devnull otherwise
    output_format : str, optional
        Output format, either "csv", "json", "parquet", or "arrow". Defaults to "parquet"
    cache_dir : Path, optional
        Directory where the parquet cache files will be generated
    limit : int, optional
        Limit number of rows to output. Defaults to None (no limit).
    typ : str, optional
        Type of the node to unbind

    Returns
    -------

    """
    from xorq.expr.api import read_pyarrow_stream
    from xorq.flight.exchanger import replace_one_unbound

    span = trace.get_current_span()
    span.add_event(
        "run_unbound.params",
        {
            "expr_path": str(expr_path),
            "to_unbind_hash": str(to_unbind_hash),
            "to_unbind_tag": str(to_unbind_tag),
            "output_format": str(output_format),
        },
    )

    # Resolve build identifier
    expr_path = ensure_build_dir(expr_path)
    # Log to stderr to avoid polluting Arrow streams
    print(f"[run-unbound] Loading expression from {expr_path}", file=sys.stderr)

    # Load the expression and make it unbound
    expr = load_expr(expr_path, cache_dir=cache_dir)
    unbound_expr = expr_to_unbound(
        expr, hash=to_unbind_hash, tag=to_unbind_tag, typs=typ
    ).to_expr()

    # Read Arrow IPC from instream
    print("[run-unbound] Reading Arrow IPC from instream...", file=sys.stderr)
    # Create a connection and register the input table
    with maybe_open(instream, "rb") as stream:
        input_expr = read_pyarrow_stream(stream)
        # Replace the unbound node with the input table
        bound_expr = replace_one_unbound(unbound_expr, input_expr)

        if limit is not None:
            bound_expr = bound_expr.limit(limit)
        arbitrate_output_format(bound_expr, output_path, output_format)


@tracer.start_as_current_span("cli.unbind_and_serve_command")
def unbind_and_serve_command(
    expr_path,
    to_unbind_hash=None,
    to_unbind_tag=None,
    host=None,
    port=None,
    prometheus_port=None,
    cache_dir=get_xorq_cache_dir(),
    typ=None,
):
    # Preserve original target token for server listing
    orig_target = expr_path
    # Resolve build identifier (alias, entry_id, build_id, or path) to an actual build directory
    expr_path = ensure_build_dir(expr_path)
    logger.info(f"Loading expression from {expr_path}")
    try:
        # initialize console and optional Prometheus metrics
        from xorq.flight.metrics import setup_console_metrics

        setup_console_metrics(prometheus_port=prometheus_port)
    except ImportError:
        logger.warning(
            "Metrics support requires 'opentelemetry-sdk' and console exporter"
        )

    expr = load_expr(expr_path, cache_dir=cache_dir)
    unbound_expr = expr_to_unbound(
        expr,
        hash=to_unbind_hash,
        tag=to_unbind_tag,
        typs=typ,
        strategy=SnapshotStrategy(),
    )
    flight_url = xorq.flight.FlightUrl(host=host, port=port)
    make_server = partial(
        xorq.flight.FlightServer,
        flight_url=flight_url,
    )
    logger.info(f"Serving expression from '{expr_path}' on {flight_url.to_location()}")
    server, _ = xorq.expr.relations.flight_serve_unbound(
        unbound_expr, make_server=make_server
    )
    # Record server metadata
    rec = ServerRecord(
        pid=os.getpid(),
        command="serve-unbound",
        target=orig_target,
        port=flight_url.port,
        node_hash=to_unbind_hash,
    )
    rec.save(Path(cache_dir) / "servers")
    server.wait()


@tracer.start_as_current_span("cli.serve_command")
def serve_command(
    expr_path,
    host=None,
    port=None,
    duckdb_path=None,
    prometheus_port=None,
    cache_dir=get_xorq_cache_dir(),
):
    """
    Serve a built expression via Flight Server

    Parameters
    ----------
    expr_path : str
        Path to the expression directory (output of xorq build)
    host : str
        Host to bind Flight Server
    port : int or None
        Port to bind Flight Server (None for random)
    duckdb_path : str or None
        Path to duckdb cache DB file
    prometheus_port : int or None
        Port to connect to the prometheus server
    cache_dir : str or None
        Path to the dir to store the parquet cache files
    """

    # Preserve original target token for server listing
    orig_target = expr_path
    # Resolve build identifier (alias, entry_id, build_id, or path) to an actual build directory
    expr_path = ensure_build_dir(expr_path)
    span = trace.get_current_span()
    params = {
        "build_path": expr_path,
        "host": host,
        "port": port,
    }
    if duckdb_path is not None:
        params["duckdb_path"] = duckdb_path
    span.add_event("serve.params", params)

    expr_path = Path(expr_path)
    logger.info(f"Loading expression from {expr_path}")
    expr = load_expr(expr_path, cache_dir=cache_dir)

    db_path = Path(duckdb_path or "xorq_serve.db")  # FIXME what should be the default?

    db_path.parent.mkdir(parents=True, exist_ok=True)
    logger.info(f"Using duckdb at {db_path}")

    try:
        from xorq.flight.metrics import setup_console_metrics

        setup_console_metrics(prometheus_port=prometheus_port)
    except ImportError:
        logger.warning(
            "Metrics support requires 'opentelemetry-sdk' and console exporter"
        )

    server = FlightServer.from_udxf(
        expr,
        make_connection=partial(load_backend("duckdb").connect, str(db_path)),
        port=port,
        host=host,
    )
    # Record server metadata
    rec = ServerRecord(
        pid=os.getpid(),
        command="serve-flight-udxf",
        target=orig_target,
        port=server.flight_url.port,
    )
    rec.save(Path(cache_dir) / "servers")
    location = server.flight_url.to_location()
    logger.info(f"Serving expression '{expr_path.stem}' on {location}")
    server.serve(block=True)


@tracer.start_as_current_span("cli.init_command")
def init_command(
    path="./xorq-template",
    template=InitTemplates.default,
    branch=None,
):
    from xorq.common.utils.download_utils import download_unpacked_xorq_template

    path = download_unpacked_xorq_template(path, template, branch=branch)
    print(f"initialized xorq template `{template}` to {path}")
    return path


def parse_args(override=None):
    parser = argparse.ArgumentParser(
        description="xorq - build, run, and serve expressions"
    )
    parser.add_argument("--pdb", action="store_true", help="Drop into pdb on failure")
    parser.add_argument(
        "--pdb-runcall", action="store_true", help="Invoke with pdb.runcall"
    )

    subparsers = parser.add_subparsers(dest="command", help="Command to execute")
    subparsers.required = True

    uv_build_parser = subparsers.add_parser(
        "uv-build", help="Build an expression with a custom Python environment"
    )
    uv_build_parser.add_argument("script_path", help="Path to the Python script")
    uv_build_parser.add_argument(
        "-e",
        "--expr-name",
        default="expr",
        help="Name of the expression variable in the Python script",
    )
    uv_build_parser.add_argument(
        "--builds-dir", default="builds", help="Directory for all generated artifacts"
    )
    uv_build_parser.add_argument(
        "--cache-dir",
        required=False,
        default=get_xorq_cache_dir(),
        help="Directory for all generated parquet files cache",
    )

    uv_run_parser = subparsers.add_parser(
        "uv-run", help="Run an expression with a custom Python environment"
    )
    uv_run_parser.add_argument("build_path", help="Path to the build script")
    uv_run_parser.add_argument(
        "--cache-dir",
        required=False,
        default=get_xorq_cache_dir(),
        help="Directory for all generated parquet files cache",
    )
    uv_run_parser.add_argument(
        "-o",
        "--output-path",
        default=None,
        help=f"Path to write output (default: {os.devnull})",
    )
    uv_run_parser.add_argument(
        "-f",
        "--format",
        choices=OutputFormats,
        default=OutputFormats.default,
        type=OutputFormats,
        help="Output format (default: parquet)",
    )

    build_parser = subparsers.add_parser(
        "build", help="Generate artifacts from an expression"
    )
    build_parser.add_argument("script_path", help="Path to the Python script")
    build_parser.add_argument(
        "-e",
        "--expr-name",
        default="expr",
        help="Name of the expression variable in the Python script",
    )
    build_parser.add_argument(
        "--builds-dir", default="builds", help="Directory for all generated artifacts"
    )
    build_parser.add_argument(
        "--cache-dir",
        required=False,
        default=get_xorq_cache_dir(),
        help="Directory for all generated parquet files cache",
    )
    build_parser.add_argument(
        "--debug",
        action="store_true",
        help="Output SQL files and other debug artifacts",
    )

    run_parser = subparsers.add_parser(
        "run", help="Run a build from a builds directory"
    )
    run_parser.add_argument("build_path", help="Path to the build script")
    run_parser.add_argument(
        "--cache-dir",
        required=False,
        default=get_xorq_cache_dir(),
        help="Directory for all generated parquet files cache",
    )
    run_parser.add_argument(
        "-o",
        "--output-path",
        default=None,
        help=f"Path to write output (default: {os.devnull})",
    )
    run_parser.add_argument(
        "-f",
        "--format",
        choices=OutputFormats,
        default=OutputFormats.default,
        type=OutputFormats,
        help="Output format (default: parquet)",
    )
    run_parser.add_argument(
        "--limit",
        type=int,
        default=None,
        help="Limit number of rows to output",
    )

    run_unbound_parser = subparsers.add_parser(
        "run-unbound", help="Run an unbound expr by reading Arrow IPC from stdin"
    )
    run_unbound_parser.add_argument(
        "build_path",
        help="Build target: alias, entry_id, build_id, or path to build dir",
    )
    run_unbound_parser.add_argument(
        "--to_unbind_hash", default=None, help="Hash of the node to unbind"
    )
    run_unbound_parser.add_argument(
        "--to_unbind_tag", default=None, help="Tag of the node to unbind"
    )
    run_unbound_parser.add_argument(
        "--typ",
        required=False,
        default=None,
        help="Type of the node to unbind",
    )
    run_unbound_parser.add_argument(
        "-o",
        "--output-path",
        default=None,
        help=f"Path to write output (default: stdout for arrow, {os.devnull} otherwise)",
    )
    run_unbound_parser.add_argument(
        "-f",
        "--format",
        choices=OutputFormats,
        # why was this arrow before and why did we fail when it was arrow?
        default=OutputFormats.default,
        type=OutputFormats,
        help=f"Output format (default: {OutputFormats.default})",
    )
    run_unbound_parser.add_argument(
        "--limit",
        type=int,
        default=None,
        help="Limit number of rows to output",
    )
    run_unbound_parser.add_argument(
        "--batch-size",
        type=int,
        default=None,
        help="Batch size for Arrow streaming output (default: use table default)",
    )
    run_unbound_parser.add_argument(
        "--cache-dir",
        required=False,
        default=get_xorq_cache_dir(),
        help="Directory for all generated parquet files cache",
    )
    run_unbound_parser.add_argument(
        "-i",
        "--instream",
        default=sys.stdin.buffer,
        help="Stream to record batches from",
    )

    serve_unbound_parser = subparsers.add_parser(
        "serve-unbound", help="Serve an an unbound expr via Flight Server"
    )
    serve_unbound_parser.add_argument(
        "build_path",
        help="Build target: alias, entry_id, build_id, or path to build dir",
    )
    serve_unbound_parser.add_argument(
        "--to_unbind_hash", default=None, help="hash of the expr to replace"
    )
    serve_unbound_parser.add_argument(
        "--to_unbind_tag", default=None, help="tag of the expr to replace"
    )
    serve_unbound_parser.add_argument(
        "--host",
        default="localhost",
        help="Host to bind Flight Server (default: localhost)",
    )
    serve_unbound_parser.add_argument(
        "--port",
        type=int,
        default=None,
        help="Port to bind Flight Server (default: random)",
    )
    serve_unbound_parser.add_argument(
        "--cache-dir",
        required=False,
        default=get_xorq_cache_dir(),
        help="Directory for all generated parquet files cache",
    )
    serve_unbound_parser.add_argument(
        "--typ",
        required=False,
        default=None,
        help="type of the node to unbind",
    )
    serve_unbound_parser.add_argument(
        "--prometheus-port",
        type=int,
        default=None,
        help="Port to expose Prometheus metrics (default: disabled)",
    )
    serve_parser = subparsers.add_parser(
        "serve-flight-udxf", help="Serve a build via Flight Server"
    )
    serve_parser.add_argument(
        "build_path",
        help="Build target: alias, entry_id, build_id, or path to build dir",
    )
    serve_parser.add_argument(
        "--host",
        default="localhost",
        help="Host to bind Flight Server (default: localhost)",
    )
    serve_parser.add_argument(
        "--port",
        type=int,
        default=None,
        help="Port to bind Flight Server (default: random)",
    )
    serve_parser.add_argument(
        "--duckdb-path",
        default=None,
        help="Path to duckdb DB (default: <build_path>/xorq_serve.db)",
    )
    serve_parser.add_argument(
        "--prometheus-port",
        type=int,
        default=None,
        help="Port to expose Prometheus metrics (default: disabled)",
    )
    serve_parser.add_argument(
        "--cache-dir",
        required=False,
        default=get_xorq_cache_dir(),
        help="Directory for all generated parquet files cache",
    )

    init_parser = subparsers.add_parser(
        "init",
        help="Initialize a xorq project",
    )
    init_parser.add_argument(
        "-p",
        "--path",
        type=Path,
        default="./xorq-template",
    )
    init_parser.add_argument(
        "-t",
        "--template",
        choices=tuple(InitTemplates),
        default=InitTemplates.cached_fetcher,
    )
    init_parser.add_argument(
        "-b",
        "--branch",
        default=None,
    )
    lineage_parser = subparsers.add_parser(
        "lineage",
        help="Print lineage trees of all columns for a build",
    )
    lineage_parser.add_argument(
        "target",
        help="Build target: alias, entry_id, build_id, or path to build dir",
    )
    ps_parser = subparsers.add_parser(
        "ps",
        help="List running xorq servers",
    )
    ps_parser.add_argument(
        "--cache-dir",
        required=False,
        default=get_xorq_cache_dir(),
        help="Directory for server state records",
    )
    catalog_parser = subparsers.add_parser("catalog", help="Manage build catalog")
    catalog_subparsers = catalog_parser.add_subparsers(
        dest="subcommand", help="Catalog commands"
    )
    catalog_subparsers.required = True

    catalog_add = catalog_subparsers.add_parser(
        "add", help="Add a build to the catalog"
    )
    catalog_add.add_argument("build_path", help="Path to the build directory")
    catalog_add.add_argument(
        "-a", "--alias", help="Optional alias for this entry", default=None
    )
    catalog_subparsers.add_parser("ls", help="List catalog entries")

    catalog_subparsers.add_parser("info", help="Show catalog information")
    catalog_rm = catalog_subparsers.add_parser(
        "rm", help="Remove a build entry or alias from the catalog"
    )
    catalog_rm.add_argument("entry", help="Entry ID or alias to remove")
    catalog_diff_builds = catalog_subparsers.add_parser(
        "diff-builds", help="Compare two build artifacts via git diff --no-index"
    )
    catalog_diff_builds.add_argument(
        "left",
        help="Left build target: alias, entry_id, build_id, or path to build dir",
    )
    catalog_diff_builds.add_argument(
        "right",
        help="Right build target: alias, entry_id, build_id, or path to build dir",
    )
    catalog_diff_builds.add_argument(
        "--all",
        action="store_true",
        help="Diff all known build files plus all .sql files",
    )
    catalog_diff_builds.add_argument(
        "--files",
        nargs="+",
        help="Explicit list of relative files to diff (overrides --all)",
        default=None,
    )

    args = parser.parse_args(override)
    return args


def main():
    """Main entry point for the xorq CLI."""
    args = parse_args()

    try:
        match args.command:
            case "uv-build":
                sys_argv = tuple(el if el != "uv-build" else "build" for el in sys.argv)
                f, f_args = (
                    uv_build_command,
                    (args.script_path, None, sys_argv),
                )
            case "uv-run":
                sys_argv = tuple(el if el != "uv-run" else "run" for el in sys.argv)
                f, f_args = (
                    uv_run_command,
                    (args.build_path, sys_argv),
                )
            case "build":
                f, f_args = (
                    build_command,
                    (
                        args.script_path,
                        args.expr_name,
                        args.builds_dir,
                        args.cache_dir,
                        args.debug,
                    ),
                )
            case "run":
                f, f_args = (
                    run_command,
                    (
                        args.build_path,
                        args.output_path,
                        args.format,
                        args.cache_dir,
                        args.limit,
                    ),
                )
            case "run-unbound":
                f, f_args = (
                    run_unbound_command,
                    (
                        args.build_path,
                        args.to_unbind_hash,
                        args.to_unbind_tag,
                        args.output_path,
                        args.format,
                        args.cache_dir,
                        args.limit,
                        args.typ,
                        args.instream,
                    ),
                )
            case "serve-unbound":
                f, f_args = (
                    unbind_and_serve_command,
                    (
                        args.build_path,
                        args.to_unbind_hash,
                        args.to_unbind_tag,
                        args.host,
                        args.port,
                        args.prometheus_port,
                        args.cache_dir,
                        args.typ,
                    ),
                )
            case "serve-flight-udxf":
                # Serve a Flight UDXF build
                f, f_args = (
                    serve_command,
                    (
                        args.build_path,
                        args.host,
                        args.port,
                        args.duckdb_path,
                        args.prometheus_port,
                        args.cache_dir,
                    ),
                )
            case "init":
                f, f_args = (
                    init_command,
                    (args.path, args.template, args.branch),
                )
            case "lineage":
                f, f_args = (
                    lineage_command,
                    (args.target,),
                )
            case "catalog":
                f, f_args = (
                    catalog_command,
                    (args,),
                )
            case "ps":
                f, f_args = (
                    ps_command,
                    (args.cache_dir,),
                )
            case _:
                raise ValueError(f"Unknown command: {args.command}")
        match args.pdb_runcall:
            case True:
                pdb.runcall(f, *f_args)
            case False:
                f(*f_args)
            case _:
                raise ValueError(f"Unknown value for pdb_runcall: {args.pdb_runcall}")
    except Exception as e:
        if args.pdb:
            traceback.print_exception(e)
            pdb.post_mortem(e.__traceback__)
        else:
            traceback.print_exc()
        sys.exit(1)


if __name__ == "__main__":
    main()
